from types import FunctionType, MethodType
from typing import Any, Optional, Type, Union

from airflow_pydantic import CallablePath, Host, HostQuery, ImportPath, Port, PortQuery, Task, TaskArgs, get_import_path
from pydantic import Field, TypeAdapter, field_validator, model_validator

from .supervisor_ssh import SupervisorSSHAirflowConfiguration

__all__ = (
    "SupervisorSSHOperatorArgs",
    "SupervisorSSHTaskArgs",
    "SupervisorSSHOperator",
    "SupervisorSSHTask",
)


class SupervisorSSHTaskArgs(TaskArgs, extra="allow"):
    cfg: SupervisorSSHAirflowConfiguration
    host: Optional[Union[Host, HostQuery, CallablePath]] = Field(
        default=None,
        description="The host to connect to for SSH, if not otherwise provided in configs",
    )
    host_foo: Optional[CallablePath] = Field(default=None, exclude=True)
    port: Optional[Union[Port, PortQuery, CallablePath]] = Field(
        default=None,
        description="The port to user for Supervisor, if not otherwise provided in configs",
    )
    port_foo: Optional[CallablePath] = Field(default=None, exclude=True)

    @model_validator(mode="before")
    @classmethod
    def _extract_host(cls, data: Any) -> Any:
        if isinstance(data, dict):
            if "host" in data:
                host = data["host"]
                if isinstance(host, (HostQuery, Host)):
                    if isinstance(host, HostQuery):
                        # Ensure that the BalancerHostQueryConfiguration is of kind 'select'
                        if not host.kind == "select":
                            raise ValueError("BalancerHostQueryConfiguration must be of kind 'select'")
                        data["host"] = host.execute()

                if isinstance(host, str):
                    # If host is a string, we assume it's an import path
                    data["host"] = get_import_path(host)

                    try:
                        data["host"] = data["host"]()
                    except Exception:
                        # Skip, might only run in situ
                        data["host"] = None

                if isinstance(host, (FunctionType, MethodType)):
                    # If host is a callable, we need to call it to get the Host instance
                    data["host_foo"] = get_import_path(host)

                    try:
                        data["host"] = data["host_foo"]()
                    except Exception:
                        # Skip, might only run in situ
                        data["host"] = None
            if "port" in data:
                port = data["port"]
                if isinstance(port, (PortQuery, Port)):
                    if isinstance(port, PortQuery):
                        # Ensure that the BalancerPortQueryConfiguration is of kind 'select'
                        if not port.kind == "select":
                            raise ValueError("BalancerPortQueryConfiguration must be of kind 'select'")
                        data["port"] = port.execute()

                if isinstance(port, str):
                    # If port is a string, we assume it's an import path
                    data["port"] = get_import_path(port)

                    try:
                        data["port"] = data["port"]()
                    except Exception:
                        # Skip, might only run in situ
                        data["port"] = None

                if isinstance(port, (FunctionType, MethodType)):
                    # If port is a callable, we need to call it to get the Port instance
                    data["port_foo"] = get_import_path(port)

                    try:
                        data["port"] = data["port_foo"]()
                    except Exception:
                        # Skip, might only run in situ
                        data["port"] = None
        return data

    @field_validator("host", mode="before")
    @classmethod
    def _validate_host(cls, v):
        if v:
            if isinstance(v, str):
                v = get_import_path(v)

            if isinstance(v, (FunctionType, MethodType)):
                try:
                    # If it's a callable, we need to call it to get the Host instance
                    v = v()
                except Exception:
                    # Skip, might only run in situ
                    v = None

            if isinstance(v, HostQuery):
                if not v.kind == "select":
                    raise ValueError("BalancerHostQueryConfiguration must be of kind 'select'")
                v = v.execute()

            if isinstance(v, dict):
                v = TypeAdapter(Host).validate_python(v)
            assert v is None or isinstance(v, Host), f"host must be an instance of Host, got: {type(v)}"
        return v

    @field_validator("port", mode="before")
    @classmethod
    def _validate_port(cls, v):
        if v:
            if isinstance(v, str):
                v = get_import_path(v)

            if isinstance(v, (FunctionType, MethodType)):
                try:
                    # If it's a callable, we need to call it to get the Port instance
                    v = v()
                except Exception:
                    # Skip, might only run in situ
                    v = None

            if isinstance(v, PortQuery):
                if not v.kind == "select":
                    raise ValueError("BalancerPortQueryConfiguration must be of kind 'select'")
                v = v.execute()

            if isinstance(v, dict):
                v = TypeAdapter(Port).validate_python(v)
            assert v is None or isinstance(v, Port), f"port must be an instance of Port, got: {type(v)}"
        return v


# Alias
SupervisorSSHOperatorArgs = SupervisorSSHTaskArgs


class SupervisorSSHTask(Task, SupervisorSSHTaskArgs):
    operator: ImportPath = Field(default="airflow_supervisor.SupervisorSSH", validate_default=True)

    @field_validator("operator")
    @classmethod
    def validate_operator(cls, v: Type) -> Type:
        if not isinstance(v, Type) and issubclass(v, SupervisorSSHAirflowConfiguration):
            raise ValueError(f"operator must be 'airflow_supervisor.SupervisorSSH', got: {v}")
        return v


# Alias
SupervisorSSHOperator = SupervisorSSHTask



from typing import Type

from airflow_pydantic import ImportPath, Task, TaskArgs
from pydantic import Field, field_validator

from .supervisor import SupervisorAirflowConfiguration

__all__ = (
    "SupervisorOperatorArgs",
    "SupervisorTaskArgs",
    "SupervisorOperator",
    "SupervisorTask",
)


class SupervisorTaskArgs(TaskArgs):
    cfg: SupervisorAirflowConfiguration


# Alias
SupervisorOperatorArgs = SupervisorTaskArgs


class SupervisorTask(Task, SupervisorTaskArgs):
    operator: ImportPath = Field(default="airflow_supervisor.Supervisor", validate_default=True)

    @field_validator("operator")
    @classmethod
    def validate_operator(cls, v: Type) -> Type:
        if not isinstance(v, Type) and issubclass(v, SupervisorAirflowConfiguration):
            raise ValueError(f"operator must be 'airflow_supervisor.Supervisor', got: {v}")
        return v


# Alias
SupervisorOperator = SupervisorTask



from airflow.models.dag import DAG


class TestSSH:
    def test_ssh_setup(self, supervisor_airflow_ssh_configuration):
        from airflow_supervisor import SupervisorSSHAirflowConfiguration

        assert supervisor_airflow_ssh_configuration is not None
        assert isinstance(supervisor_airflow_ssh_configuration, SupervisorSSHAirflowConfiguration)
        assert supervisor_airflow_ssh_configuration.ssh_operator_args.ssh_hook.remote_host == "localhost"

    def test_ssh_overrides(self, supervisor_airflow_ssh_configuration):
        from airflow_pydantic import Host

        from airflow_supervisor import SupervisorSSH

        dag = DAG(dag_id="test_dag", default_args={}, schedule=None, params={})
        inst = SupervisorSSH(
            dag=dag,
            cfg=supervisor_airflow_ssh_configuration,
            host=Host(name="test-host"),
        )
        assert inst._cfg.ssh_operator_args.ssh_hook.remote_host == "test-host"



from getpass import getuser
from pathlib import Path
from unittest.mock import patch

import pytest
from airflow.models.pool import Pool
from airflow_balancer.testing import pools, variables
from airflow_config import load_config
from airflow_pydantic import SSHOperatorArgs


def test_hydra_config():
    try:
        from airflow_supervisor import SupervisorSSHAirflowConfiguration
    except ImportError:
        pytest.skip("SupervisorSSHAirflowConfiguration not available")

    with (
        patch("supervisor_pydantic.config.supervisor.gettempdir") as p1,
    ):
        with pools():
            pth = Path(__file__).resolve().parent.parent.parent.parent / ".pytest_cache"
            p1.return_value = str(pth)
            cfg = load_config("config", "config")
            assert "balancer" in cfg.extensions
            assert "supervisor" in cfg.extensions

            supervisor_cfg: SupervisorSSHAirflowConfiguration = cfg.extensions["supervisor"]

            # Basic Supervisor checks
            assert supervisor_cfg.command_prefix == ""
            assert supervisor_cfg.working_dir == pth / f"supervisor-{getuser()}-sleep-echo"
            assert supervisor_cfg.inet_http_server.port == "*:9001"
            assert str(supervisor_cfg.supervisorctl.serverurl) == "http://localhost:9001/"

            assert "sleep" in supervisor_cfg.program
            assert "echo" in supervisor_cfg.program
            assert supervisor_cfg.program["sleep"].command == "sleep 1000"
            assert supervisor_cfg.program["echo"].command == 'echo "hello"'

            # SSH Operator checks
            ssh_args: SSHOperatorArgs = supervisor_cfg.ssh_operator_args
            assert ssh_args is not None
            assert ssh_args.cmd_timeout == 63


def test_hydra_config_render():
    with (
        patch("supervisor_pydantic.config.supervisor.gettempdir") as p1,
        patch("supervisor_pydantic.config.supervisor.getuser") as p2,
    ):
        with pools():
            pth = Path(__file__).resolve().parent.parent.parent.parent / ".pytest_cache"
            p1.return_value = str(pth)
            p2.return_value = "user"
            cfg = load_config("config", "config_tasks")

            dag = cfg.dags["example_dag"]
            assert (
                dag.render()
                == """# Generated by airflow-config
from datetime import datetime, time, timedelta
from pathlib import Path

from airflow.models import DAG

from airflow_supervisor.airflow.ssh import SupervisorSSH

with DAG(
    schedule="0 3 * * *",
    start_date=datetime.fromisoformat("2023-01-01T00:00:00"),
    dag_id="example_dag",
    default_args={},
) as dag:
    run = SupervisorSSH(
        cfg={
            "inet_http_server": {"port": "*:9001", "username": None, "password": None},
            "program": {
                "echo": {
                    "command": 'echo "hello"',
                    "autostart": False,
                    "startsecs": 1,
                    "startretries": None,
                    "autorestart": False,
                    "exitcodes": [0],
                    "stopsignal": "TERM",
                    "stopwaitsecs": 30,
                    "stopasgroup": True,
                    "killasgroup": True,
                    "stdout_logfile": Path("/data/echo/output.log"),
                    "stderr_logfile": Path("/data/echo/error.log"),
                    "directory": Path("/data/echo"),
                },
                "sleep": {
                    "command": "sleep 1000",
                    "autostart": False,
                    "startsecs": 1,
                    "startretries": None,
                    "autorestart": False,
                    "exitcodes": [0],
                    "stopsignal": "TERM",
                    "stopwaitsecs": 30,
                    "stopasgroup": True,
                    "killasgroup": True,
                    "stdout_logfile": Path("/data/sleep/output.log"),
                    "stderr_logfile": Path("/data/sleep/error.log"),
                    "directory": Path("/data/sleep"),
                },
            },
            "rpcinterface": {"supervisor": {"rpcinterface_factory": "supervisor.rpcinterface:make_main_rpcinterface"}},
            "config_path": Path("/data/supervisord.conf"),
            "working_dir": Path("/data"),
            "check_interval": timedelta(10.0),
            "check_timeout": timedelta(60.0),
            "runtime": timedelta(360.0),
            "endtime": time(23, 0, 0, 0),
            "maxretrigger": 5,
            "stop_on_exit": False,
            "cleanup": False,
            "restart_on_initial": True,
            "restart_on_retrigger": False,
            "ssh_operator_args": {"cmd_timeout": 63},
        },
        task_id="run",
        dag=dag,
    )
"""
            )

        exec(dag.render())


def test_hydra_config_render_hosts():
    with (
        patch("supervisor_pydantic.config.supervisor.gettempdir") as p1,
        patch("supervisor_pydantic.config.supervisor.getuser") as p2,
    ):
        with pools():
            pth = Path(__file__).resolve().parent.parent.parent.parent / ".pytest_cache"
            p1.return_value = str(pth)
            p2.return_value = "user"
            cfg = load_config("config", "config_tasks_host")

            dag = cfg.dags["example_dag"]
            assert (
                dag.render()
                == """# Generated by airflow-config
from datetime import datetime
from pathlib import Path

from airflow.models import DAG
from airflow_pydantic import Host, Port

from airflow_supervisor.airflow.ssh import SupervisorSSH

with DAG(
    schedule="0 3 * * *",
    start_date=datetime.fromisoformat("2023-01-01T00:00:00"),
    dag_id="example_dag",
    default_args={},
) as dag:
    run = SupervisorSSH(
        cfg={
            "inet_http_server": {"port": "*:9001", "username": None, "password": None},
            "program": {
                "echo": {
                    "command": 'echo "hello"',
                    "autostart": False,
                    "startsecs": 1,
                    "startretries": None,
                    "autorestart": False,
                    "exitcodes": [0],
                    "stopsignal": "TERM",
                    "stopwaitsecs": 30,
                    "stopasgroup": True,
                    "killasgroup": True,
                    "stdout_logfile": Path("/data/echo/output.log"),
                    "stderr_logfile": Path("/data/echo/error.log"),
                    "directory": Path("/data/echo"),
                }
            },
            "rpcinterface": {"supervisor": {"rpcinterface_factory": "supervisor.rpcinterface:make_main_rpcinterface"}},
            "config_path": Path("/data/supervisord.conf"),
            "working_dir": Path("/data"),
        },
        host=Host(name="my-remote-host", password_variable="blerg", password_variable_key="password", os="rhel"),
        port=Port(
            name="my-remote-port",
            host=Host(name="my-remote-host", password_variable="blerg", password_variable_key="password", os="rhel"),
            port=8080,
        ),
        task_id="run",
        dag=dag,
    )
"""
            )

        exec(dag.render())


def test_hydra_config_render_hosts_query():
    with (
        patch("supervisor_pydantic.config.supervisor.gettempdir") as p1,
        patch("supervisor_pydantic.config.supervisor.getuser") as p2,
    ):
        with pools():
            pth = Path(__file__).resolve().parent.parent.parent.parent / ".pytest_cache"
            p1.return_value = str(pth)
            p2.return_value = "user"
            cfg = load_config("config", "config_tasks_host_query")

            dag = cfg.dags["example_dag"]
            assert (
                dag.render()
                == """# Generated by airflow-config
from datetime import datetime
from pathlib import Path

from airflow.models import DAG
from airflow.models.pool import Pool
from airflow_pydantic import Host, Port

from airflow_supervisor.airflow.ssh import SupervisorSSH

with DAG(
    schedule="0 3 * * *",
    start_date=datetime.fromisoformat("2023-01-01T00:00:00"),
    dag_id="example_dag",
    default_args={},
) as dag:
    run = SupervisorSSH(
        cfg={
            "inet_http_server": {"port": "*:9001", "username": None, "password": None},
            "program": {
                "echo": {
                    "command": 'echo "hello"',
                    "autostart": False,
                    "startsecs": 1,
                    "startretries": None,
                    "autorestart": False,
                    "exitcodes": [0],
                    "stopsignal": "TERM",
                    "stopwaitsecs": 30,
                    "stopasgroup": True,
                    "killasgroup": True,
                    "stdout_logfile": Path("/data/echo/output.log"),
                    "stderr_logfile": Path("/data/echo/error.log"),
                    "directory": Path("/data/echo"),
                }
            },
            "rpcinterface": {"supervisor": {"rpcinterface_factory": "supervisor.rpcinterface:make_main_rpcinterface"}},
            "config_path": Path("/data/supervisord.conf"),
            "working_dir": Path("/data"),
        },
        host=Host(
            name="server2",
            username="user1",
            password_variable="myvar",
            password_variable_key="password",
            pool=Pool.create_or_update_pool(
                name="server2", slots=8, description="Balancer pool for host(server2)", include_deferred=False
            ).pool,
            size=8,
            tags=["tag2"],
        ),
        port=Port(
            name="my-app",
            host=Host(
                name="server2",
                username="user1",
                password_variable="myvar",
                password_variable_key="password",
                pool=Pool.create_or_update_pool(
                    name="server2", slots=8, description="Balancer pool for host(server2)", include_deferred=False
                ).pool,
                size=8,
                tags=["tag2"],
            ),
            host_name="server2",
            port=8080,
        ),
        task_id="run",
        dag=dag,
    )
"""
            )
        with pools(Pool()), variables({"username": "user1", "password": "myvar"}):
            exec(dag.render())



from typing import Optional

from airflow_pydantic import SSHOperatorArgs
from pydantic import Field
from supervisor_pydantic import SupervisorLocation

from .supervisor import SupervisorAirflowConfiguration

__all__ = (
    "SupervisorSSHAirflowConfiguration",
    "SSHOperatorArgs",  # reexport
    "load_airflow_ssh_config",
)


class SupervisorSSHAirflowConfiguration(SupervisorAirflowConfiguration):
    command_prefix: Optional[str] = Field(default="")

    ssh_operator_args: Optional[SSHOperatorArgs] = Field(
        default=None,
        description="SSH Operator arguments to use for remote execution.",
    )

    # NOTE: Override in base-base class
    local_or_remote: Optional[SupervisorLocation] = Field(
        default="remote",
        description="Location of supervisor, either local for same-machine or remote. If same-machine, communicates via Unix sockets by default, if remote, communicates via inet http server",
    )


load_airflow_ssh_config = SupervisorSSHAirflowConfiguration.load



from getpass import getuser
from pathlib import Path
from unittest.mock import patch

import pytest

from airflow_supervisor import load_airflow_config, load_airflow_ssh_config

EXPECTED_CFG = """[inet_http_server]
port=*:9001

[supervisord]
logfile={dir}/supervisord.log
pidfile={dir}/supervisord.pid
nodaemon=false
directory={dir}
identifier=supervisor

[supervisorctl]
serverurl=http://localhost:9001/

[program:sleep]
command=sleep 1000
autostart=false
startsecs=1
autorestart=false
exitcodes=0
stopsignal=TERM
stopwaitsecs=30
stopasgroup=true
killasgroup=true
stdout_logfile={dir}/sleep/output.log
stderr_logfile={dir}/sleep/error.log
directory={dir}/sleep

[program:echo]
command=echo "hello"
autostart=false
startsecs=1
autorestart=false
exitcodes=0
stopsignal=TERM
stopwaitsecs=30
stopasgroup=true
killasgroup=true
stdout_logfile={dir}/echo/output.log
stderr_logfile={dir}/echo/error.log
directory={dir}/echo

[rpcinterface:supervisor]
supervisor.rpcinterface_factory=supervisor.rpcinterface:make_main_rpcinterface"""


def test_hydra_overrides():
    with (
        patch("supervisor_pydantic.config.supervisor.gettempdir") as p1,
    ):
        pth = Path(__file__).resolve().parent.parent.parent.parent / ".pytest_cache"
        p1.return_value = str(pth)
        cfg = load_airflow_config(
            "config", overrides=["+program=[sleep,echo]", "+rpcinterface=standard", "+inet_http_server=local"]
        )
        assert cfg.to_cfg().strip() == EXPECTED_CFG.format(dir=str(pth / f"supervisor-{getuser()}-sleep-echo"))


def test_hydra_overrides_ssh():
    try:
        import airflow_supervisor.config.supervisor_ssh  # noqa: F401
    except ImportError:
        pytest.skip("airflow-balancer not available")
    with (
        patch("supervisor_pydantic.config.supervisor.gettempdir") as p1,
    ):
        pth = Path(__file__).resolve().parent.parent.parent.parent / ".pytest_cache"
        p1.return_value = str(pth)
        cfg = load_airflow_ssh_config(
            "config",
            "supervisor",
        )
        assert cfg.to_cfg().strip() == EXPECTED_CFG.format(dir=str(pth / f"supervisor-{getuser()}-sleep-echo"))



from datetime import time, timedelta
from typing import Literal, Optional

from pydantic import Field
from supervisor_pydantic import SupervisorConvenienceConfiguration

__all__ = (
    "SupervisorAirflowConfiguration",
    "load_airflow_config",
)


# PredefinedTemplates = Literal[
#     "always_on",
#     "half_day",
#     # TODO add more
# ]


class SupervisorAirflowConfiguration(SupervisorConvenienceConfiguration):
    """Settings that MUST be set when running in airflow"""

    # Passthrough to PythonSensor in airflow-ha
    check_interval: timedelta = Field(
        default=timedelta(seconds=5), description="Interval between supervisor program status checks"
    )
    check_timeout: timedelta = Field(
        default=timedelta(hours=8), description="Timeout to wait for supervisor program status checks"
    )

    # HighAvailabilityOperator custom args
    runtime: Optional[timedelta] = Field(default=None, description="Max runtime of Supervisor job")
    endtime: Optional[time] = Field(default=None, description="End time of Supervisor job")
    maxretrigger: Optional[int] = Field(
        default=None,
        description="Max number of retriggers of Supervisor job (e.g. max number of checks separated by `check_interval`)",
    )

    reference_date: Literal["start_date", "logical_date", "data_interval_end"] = Field(
        default="data_interval_end",
        description="Reference date for the job. NOTE: Airflow schedules after end of date interval, so `data_interval_end` is the default",
    )

    # Airflow-specific settings
    pool: Optional[str] = Field(
        default=None,
        description="Airflow pool to use for the job. If not set, the job will use the default pool, or the pool from a balancer host.",
    )

    """Other Airflow Configuration"""
    # Should the programs be stopped when the DAG finishes?
    stop_on_exit: Optional[bool] = Field(default=True, description="Stop supervisor on dag completion")

    # Should the supervisor folder be removed on dag completion?
    cleanup: Optional[bool] = Field(
        default=True, description="Cleanup supervisor folder on dag completion. Note: stop_on_exit must be True"
    )

    # For Jobs that do not shutdown, e.g. stop_on_exit=False, one might want to configure them to
    # restart when the DAG is rescheduled by airflow or retriggered by airflow-ha
    restart_on_initial: Optional[bool] = Field(
        default=False,
        description="Restart the job when the DAG is run directly via airflow (NOT retriggered). This is useful for jobs that do not shutdown",
    )
    restart_on_retrigger: Optional[bool] = Field(
        default=False,
        description="Restart the job when the DAG is retriggered. This is useful for jobs that do not shutdown",
    )

    # template: Optional[PredefinedTemplates] = Field(
    #     default=None,
    #     description="A template of settings to use. This is a convenience for common settings. Individual settings will override the template.",
    # )

    # @model_validator(mode="after")
    # def _set_fields_from_template(self):
    #     if self.template == "always_on":

    #         self.stop_on_exit = False
    #         self.cleanup = False
    #         self.restart_on_initial = True
    #         self.restart_on_retrigger = True
    #     elif self.template == "half_day":
    #         self.runtime = self.runtime or self.timedelta(hours=12)
    #         self.stop_on_exit = True
    #         self.cleanup = True
    #         self.restart_on_initial = False
    #         self.restart_on_retrigger = False
    #     return self


load_airflow_config = SupervisorAirflowConfiguration.load
